﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Xml;
using System.Text;
using System.Reflection;
using System.Collections;
using System.Collections.Generic;

namespace XAMLTools
{
    public class XamlWriter
    {
        public static string Write(DependencyObject parent)
        {
            StringBuilder sb = new StringBuilder();
            XmlWriterSettings settings = new XmlWriterSettings();
            settings.OmitXmlDeclaration = true;
            settings.Indent = true;
            XmlWriter writer = XmlWriter.Create(sb, settings);
            if (!TypeValidator.isValid(parent))
                throw new TypeValidationException();
            writer.WriteStartElement(parent.GetType().Name);
            writeCLRProperties(parent, writer);
            writeDPs(parent, writer);
            writeBrushes(parent, writer);
            writeCollections(parent, writer);
            writer.WriteEndElement();
            writer.Flush();
            writer.Close();
            return XamlHelper.getSilverlightCompatibleXaml(sb.ToString());
        }
        private static void writeCLRProperties(DependencyObject target, XmlWriter writer)
        {
            PropertyInfo[] props = target.GetType().GetProperties();
            bool hdc = false;
            foreach (var prop in props)
            {
                hdc = XamlHelper.hasDeepCollections(target, prop);
                if (prop.Name == "Name")
                {
                    if (prop.GetValue(target, null).ToString() != "")
                        writer.WriteAttributeString("x" + prop.Name, prop.GetValue(target, null).ToString());
                }
                else if (prop.Name != "Parent" && prop.CanRead && prop.CanWrite && prop.GetValue(target, null) != null
                    && prop.PropertyType.BaseType.Name != "Collection`1" && prop.PropertyType.Name != "Brush"
                    && !hdc)
                    writer.WriteAttributeString(prop.Name, prop.GetValue(target, null).ToString());
            }
        }
        private static void writeDPs(DependencyObject target, XmlWriter writer)
        {
            //DPs problem
            try
            {
                writer.WriteAttributeString("Canvas.Left", target.GetValue(Canvas.LeftProperty).ToString());
                writer.WriteAttributeString("Canvas.Top", target.GetValue(Canvas.TopProperty).ToString());
                writer.WriteAttributeString("Canvas.ZIndex", target.GetValue(Canvas.ZIndexProperty).ToString());
            }
            catch { }
            try
            {
                writer.WriteAttributeString("Storyboard.TargetName", target.GetValue(Storyboard.TargetNameProperty).ToString());
                writer.WriteAttributeString("Storyboard.TargetProperty", target.GetValue(Storyboard.TargetPropertyProperty).ToString());
            }
            catch { }

        }
        private static void writeBrushes(DependencyObject target, XmlWriter writer)
        {
            PropertyInfo[] props = target.GetType().GetProperties();
            object val;
            foreach (var prop in props)
                if (prop.PropertyType.Name == "Brush" && (val = prop.GetValue(target, null) ?? new object()).ToString() != "")
                {
                    if (val.ToString() == "System.Object")
                        continue;
                    writer.WriteStartElement(prop.ReflectedType.Name + "." + prop.Name);
                    writer.WriteStartElement(val.ToString().Split('.')[3]);
                    writeCLRProperties((DependencyObject)val, writer);
                    writeDeepCollections((DependencyObject)val, writer);
                    writeCollections((DependencyObject)val, writer);
                    writer.WriteEndElement();
                    writer.WriteEndElement();
                }
        }
        private static void writeCollections(DependencyObject target, XmlWriter writer)
        {
            if (target == null)
                return;
            PropertyInfo[] props = target.GetType().GetProperties();
            int cnt;
            object val;
            string collectionElement = "";
            foreach (var prop in props)
                if (prop.Name != "Parent" && prop.PropertyType.BaseType != null && prop.PropertyType.BaseType.Name == "Collection`1")
                {
                    cnt = (int)prop.PropertyType.InvokeMember("get_Count", BindingFlags.InvokeMethod, null, prop.GetValue(target, null), null);
                    for (int i = 0; i < cnt; i++)
                    {
                        val = prop.PropertyType.InvokeMember("get_Item", BindingFlags.InvokeMethod, null, prop.GetValue(target, null), new object[] { i });
                        if (XamlHelper.hasCollections((DependencyObject)val))
                        {
                            if (collectionElement != prop.ReflectedType.Name + "." + prop.Name)
                            {
                                if (collectionElement != "")
                                    writer.WriteEndElement();
                                collectionElement = prop.ReflectedType.Name + "." + prop.Name;
                                writer.WriteStartElement(collectionElement);
                            }
                            writer.WriteStartElement(val.GetType().Name);
                            writeCLRProperties((DependencyObject)val, writer);
                            writeDPs((DependencyObject)val, writer);
                            writeBrushes((DependencyObject)val, writer);
                            writeCollections((DependencyObject)val, writer);
                            writer.WriteEndElement();
                        }
                        else
                        {
                            writer.WriteStartElement(val.GetType().Name);
                            writeCLRProperties((DependencyObject)val, writer);
                            writeDPs((DependencyObject)val, writer);
                            writeBrushes((DependencyObject)val, writer);
                            writeDeepCollections((DependencyObject)val, writer);
                            writer.WriteEndElement();
                        }
                    }
                    if (collectionElement != "")
                    {
                        writer.WriteEndElement();
                        collectionElement = "";
                    }
                }
                //TransformGroup not inherits from Collection, so it can not be added
                //through writeCollections/writeDeepCollections and we should handle
                //it separately
                else if (prop.PropertyType.Name == "Transform")
                {
                    val = prop.GetValue(target, null);
                    if (val == null)
                        continue;
                    if (val.GetType().ToString() != "System.Windows.Media.TransformGroup" ||
                        ((int)val.GetType().GetProperty("Children").GetValue(val, null).GetType().GetProperty("Count").GetValue(val.GetType().GetProperty("Children").GetValue(val, null), null)) == 0)
                        continue;
                    writer.WriteStartElement(prop.ReflectedType.Name + "." + prop.Name);
                    writer.WriteStartElement(val.GetType().Name);
                    writeCollections(val as DependencyObject, writer);
                    writer.WriteEndElement();
                    writer.WriteEndElement();
                }
        }
        private static void writeDeepCollections(DependencyObject target, XmlWriter writer)
        {
            PropertyInfo[] props = target.GetType().GetProperties();
            object val;
            foreach (var prop in props)
                if (prop.PropertyType.BaseType.Name != "Collection`1" && prop.PropertyType.Name != "Transform" && XamlHelper.hasDeepCollections(target, prop))
                {
                    val = prop.GetValue(target, null);
                    if (val != null)
                    {
                        writer.WriteStartElement(prop.Name);
                        writeCollections((DependencyObject)val, writer);
                        writer.WriteEndElement();
                    }
                }
        }
    }
}
